let TestResult = require('./TestResults.js');
let TrendsData = require('./TrendsData.js');
let RallyProcessing = require('./Rally.js');;
let PostingController = require('./PostingController.js');
var commonFunctions = require('./commonFunctions.js')

const path = require('path');

class ParseTestData {
    constructor() {
        this.RallyProcessingObj = new RallyProcessing();
        this.PostingControllerObj = new PostingController();
    }

    async postResults(headerConfigProps, RallyJson) {
        //trim all extra spaces from attribute name and value
        let rallyDataParsingIssues = await this.parseRallyProcessingAttr(headerConfigProps);
        let trendsObj;
        if (headerConfigProps.generatetrends && (headerConfigProps.generatetrends === "true" || headerConfigProps.generatetrends === true))
            trendsObj = await this.parseTrendsAttr(headerConfigProps);
        let parsedData;
        parsedData = await this.parseTCAttributes(headerConfigProps, RallyJson);
        if (!parsedData.errors && rallyDataParsingIssues.length === 0) {
            return this.PostingControllerObj.logResults(parsedData.data, this.RallyProcessingObj, headerConfigProps.generatetrends, trendsObj);
        } else {
            if (parsedData.message)
                parsedData.message = [...parsedData.message, rallyDataParsingIssues].join("");
            else
                parsedData.message = rallyDataParsingIssues.join("");

            return parsedData;
        }
    }

    async parseRallyProcessingAttr(headerConfigProps) {
        await this.RallyProcessingObj.setRallyUserApiKey(headerConfigProps.apikey, "apikey")
        await this.RallyProcessingObj.setRallyUserName(headerConfigProps.rallyusername)
        await this.RallyProcessingObj.setRallyProxy(headerConfigProps.rallyproxy)
        if (headerConfigProps.createnewtestcases)
            await this.RallyProcessingObj.setCreateNewTestCases(headerConfigProps.createnewtestcases)
        if (headerConfigProps.updatetcr)
            await this.RallyProcessingObj.setUpdateTCR(headerConfigProps.updatetcr);
        if (this.RallyProcessingObj.invalidDataCollection)
            return this.RallyProcessingObj.invalidDataCollection;
    }

    async parseTrendsAttr(headerConfigProps) {
        let trendsObj = new TrendsData();
        await trendsObj.setTrendsModule(headerConfigProps.trendsmodule);
        await trendsObj.setTrendsEnv(headerConfigProps.trendsenv);
        await trendsObj.setTrendsJobname(headerConfigProps.trendsjobname);
        await trendsObj.setAwss3Proxy(headerConfigProps.awss3proxy);
        await trendsObj.setFrameWorkFolder(headerConfigProps.framework);
        return trendsObj;
    }

    async parseTCAttributes(headerConfigProps, RallyJson) {
        let parsingErrors = [], allFeatureResults = []
        for (let feature of RallyJson) {
            let feature_name = feature.name;
            let featureNamePath = feature.uri, featureLevelRallyAttr = null, featureLevelRallyElem = null;
            if (feature_name.includes('(**') && feature_name.includes('**)')) {
                featureLevelRallyAttr = feature_name.substring(feature_name.indexOf('(**') + 3, feature_name.indexOf('**)'));
                featureLevelRallyElem = commonFunctions.splitString(featureLevelRallyAttr, ',');
            }
            //verify priority of projectName and FolderPath at feature level itself as scenario description doesn't have these properties
            let highlevelRallyAttrVal = await this.findHighLevelRallyAttrValues(featureLevelRallyElem, headerConfigProps, featureNamePath, parsingErrors);
            //parse complete json to find all the high level error and display all at once. So, continue parsing
            if (highlevelRallyAttrVal.errors && highlevelRallyAttrVal.message && highlevelRallyAttrVal.message.length > 0) {
                parsingErrors.push(...highlevelRallyAttrVal.message);
            }
            let result;
            for (let scenario of feature.elements) {
                result = await this.assignTCProperties(scenario, featureLevelRallyElem, headerConfigProps, feature_name, featureNamePath, highlevelRallyAttrVal);
                if (result != null)
                    allFeatureResults.push(result);
            }
        }
        for (let finalResult of allFeatureResults) {
            for (let details of finalResult.invalidDataCollection) {
                if (!parsingErrors.includes(details))
                    parsingErrors.push(details);
            }
        }
        if (parsingErrors.length > 0) {
            console.error("Rally Posting Failed due to below errors : \n " + parsingErrors);
            return { "errors": true, "message": parsingErrors };
        }
        return { "errors": false, "data": allFeatureResults };
    }

    async findHighLevelRallyAttrValues(featureLevelRallyElem, headerConfigProps, featureNamepath, parsingErrors) {
        let returnResult = {};
        let errorsInAttributeValues = [];
        let preferedFolderPath = await this.findandGetPreferedData(headerConfigProps, featureLevelRallyElem, 'folderpath', 'overwrite_fp');
        if (preferedFolderPath === null) {
            let tempMessage = "Rally attribute 'folderpath' is not present in " + featureNamepath + ".\n";
            returnResult.errors = true;
            if (!parsingErrors.includes(tempMessage))// Check duplicate error message for multiple scenario in feature file
                errorsInAttributeValues.push(tempMessage);
        }
        else if (preferedFolderPath === '' || preferedFolderPath.length === 0) {
            let tempMessage = "Invalid 'folderpath' attribute value in " + featureNamepath + ".\n";
            returnResult.errors = true;
            if (!parsingErrors.includes(tempMessage))// Check duplicate error message for multiple scenario in feature file
                errorsInAttributeValues.push(tempMessage);
        }
        returnResult.folderpath = preferedFolderPath;
        let preferedProject = await this.findandGetPreferedData(headerConfigProps, featureLevelRallyElem, 'projectname', 'overwrite_pn');
        if (preferedProject === null) {
            let tempMessage = "Rally attribute 'projectname' is not present in " + featureNamepath + ".\n";
            returnResult.errors = true;
            if (!parsingErrors.includes(tempMessage))// Check duplicate error message for multiple scenario in feature file
                errorsInAttributeValues.push(tempMessage);
        }
        else if (preferedProject === '' || preferedProject.length === 0) {
            let tempMessage = "Invalid 'projectname' attribute value in " + featureNamepath + ".\n";
            returnResult.errors = true;
            if (!parsingErrors.includes(tempMessage))// Check duplicate error message for multiple scenario in feature file
                errorsInAttributeValues.push(tempMessage);
        }
        returnResult.projectname = preferedProject;
        if (errorsInAttributeValues.length > 0) { returnResult.message = errorsInAttributeValues; }
        return returnResult;
    }

    /* will assign all the rally properties that are parsed from json, if scenarioLevelRallyElements are null then will not parse anything */
    async assignTCProperties(scenarioJson, featureLevelRallyElem, headerConfigProps, feature_name, featureNamePath, highlevelRallyAttrVal) {
        let scenario_name = scenarioJson.name;
        let featureFileName = path.win32.basename(featureNamePath);
        featureFileName = featureFileName.substring(0, featureFileName.indexOf('.'));
        let scenarioLevelRallyElements = null, scenarioName, rallyAttrStr;
        if (scenario_name.includes('(**') && scenario_name.includes('**)')) {
            rallyAttrStr = scenario_name.substring(scenario_name.indexOf('(**') + 3, scenario_name.indexOf('**)'));
            scenarioName = scenario_name.substring(0, scenario_name.indexOf('(**')).trim();
            scenarioLevelRallyElements = commonFunctions.splitString(rallyAttrStr, ',');
        } else {
            scenarioName = scenario_name.trim();
        }
        let result;
        result = new TestResult();
        result.setFeatureName(feature_name.substring(0, feature_name.indexOf('(**') != -1 ? feature_name.indexOf('(**') : feature_name.length));
        result.setFeatureNamePath(featureNamePath);
        result.setTestName(scenarioName);
        result.verifySpecialCharacter(featureFileName,
            "Special characters " + await result.getSpecialChars() + " not allowed in feature file name. Special characters found for feature file " + featureNamePath + ".");
        if (featureLevelRallyElem)
            result.verifyInvalidAttributes(Object.keys(featureLevelRallyElem));
        if (scenarioLevelRallyElements) {
            result.verifyInvalidAttributes(Object.keys(scenarioLevelRallyElements));
        }
        if (!highlevelRallyAttrVal.errors) {
            result.setProjectName(highlevelRallyAttrVal.projectname);
            result.setTestFolderPath(highlevelRallyAttrVal.folderpath + "/" + featureFileName);
        }
        let testsetIdToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'testsetid', 'overwrite_ts')
        if (testsetIdToSet)
            result.setTestSetId(testsetIdToSet);
        let usToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'userstory', 'overwrite_us')
        if (usToSet)
            result.setUserStory(usToSet);
        let buildToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'build', 'overwrite_build')
        if (buildToSet === null || buildToSet === undefined) buildToSet = 1;
        result.setBuild(buildToSet);
        let tagsToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'tags', 'overwrite_tags');
        if (tagsToSet)
            result.setTags(tagsToSet);
        let methodToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'method', 'overwrite_method');
        if (methodToSet)
            result.setMethod(methodToSet);
        let typeToSet = await this.findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, 'testtype', 'overwrite_testtype');
        if (typeToSet)
            result.setTestType(typeToSet);
        if (scenarioLevelRallyElements != null) {
            if (scenarioLevelRallyElements.hasOwnProperty("description"))
                result.setDescription(scenarioLevelRallyElements["description"]);
            if (scenarioLevelRallyElements.hasOwnProperty("objective"))
                result.setObjective(scenarioLevelRallyElements["objective"]);
            if (scenarioLevelRallyElements.hasOwnProperty("priority"))
                result.setPriority(scenarioLevelRallyElements["priority"]);
            if (scenarioLevelRallyElements.hasOwnProperty("risk"))
                result.setRisk(scenarioLevelRallyElements["risk"]);
            if (scenarioLevelRallyElements.hasOwnProperty("defectid"))
                result.setDefectId(scenarioLevelRallyElements["defectid"]);
            if (scenarioLevelRallyElements.hasOwnProperty("testcaseid"))
                result.setTestCaseId(scenarioLevelRallyElements["testcaseid"]);
            if (scenarioLevelRallyElements.hasOwnProperty("preconditions"))
                result.setPreConditions(scenarioLevelRallyElements["preconditions"]);
            if (scenarioLevelRallyElements.hasOwnProperty("postconditions"))
                result.setPostConditions(scenarioLevelRallyElements["postconditions"]);
            /* if (scenarioLevelRallyElements.hasOwnProperty("tcstepinput"))
                result.setTCStepInput(scenarioLevelRallyElements["tcstepinput"]);
            if (scenarioLevelRallyElements.hasOwnProperty("tcstepexpresult"))
                result.setTCStepExpResult(scenarioLevelRallyElements["tcstepexpresult"]); */
        }
        let stepDetails = await this.getStepDetails(scenarioJson["steps"])
        result.setTestStatus(stepDetails.status);
        result.setDurationMS(Math.round(stepDetails.duration / 1000000))
        if (stepDetails.errors != null && stepDetails.errors != undefined) {
            result.setNotes(stepDetails.errors);
        }
        return result;
    }



    // limited step status to skipped, failed, passed. 
    async getStepDetails(steps) {
        let result = {
            status: null,
            duration: 0,
            errors: ''
        }
        let tempStatus = 'Pass';
        for (let step of steps) {
            if (step.result.error_message != null && step.result.error_message != undefined)
                result.errors += step.result.error_message;
            result.duration += step.result.duration;
            if (step.result.status === 'failed') {
                tempStatus = 'Fail';
                break;
            } else if (step.result.status === 'skipped' && step === steps[0]) {
                tempStatus = 'Inconclusive';
                break;
            }

        }
        result.status = tempStatus;
        return result
    }

    //verifying header and  features rally attributes alone for preference.    
    async findandGetPreferedData(headerConfigProps, featureLevelRallyElem, attrName, overwriteAttrName) {
        let preferedValue = null;
        if (!(commonFunctions.isNullEmptyOrBlank(headerConfigProps[attrName])) && (headerConfigProps[overwriteAttrName] === "true" || headerConfigProps[overwriteAttrName] === true)) {
            preferedValue = headerConfigProps[attrName];
        }
        else if (!(featureLevelRallyElem && featureLevelRallyElem.hasOwnProperty(attrName))) {
            if (!(commonFunctions.isNullEmptyOrBlank(headerConfigProps[attrName])))
                preferedValue = headerConfigProps[attrName];
        }
        else
            preferedValue = featureLevelRallyElem[attrName];
        return preferedValue;
    }


    //verifying header, feature, scenario rally attributes for preference.    
    async findandGetDeepPereferedData(headerConfigProps, featureLevelRallyElem, scenarioLevelRallyElements, attrName, overwriteAttrName) {
        let preferedValue = null;
        if (!(commonFunctions.isNullEmptyOrBlank(headerConfigProps[attrName])) && (headerConfigProps[overwriteAttrName] === "true" || headerConfigProps[overwriteAttrName] === true)) {
            preferedValue = headerConfigProps[attrName];
        }
        else if (!(scenarioLevelRallyElements && scenarioLevelRallyElements.hasOwnProperty(attrName)) && !(featureLevelRallyElem && featureLevelRallyElem.hasOwnProperty(attrName))) {
            if (!(commonFunctions.isNullEmptyOrBlank(headerConfigProps[attrName])))
                preferedValue = headerConfigProps[attrName];
        }
        else
            preferedValue = (scenarioLevelRallyElements &&scenarioLevelRallyElements.hasOwnProperty(attrName)) ? scenarioLevelRallyElements[attrName] : featureLevelRallyElem ? featureLevelRallyElem[attrName] : null;
        return preferedValue;
    }
}

module.exports = ParseTestData;