var RallyBatchAPIController = require('./RallyBatchAPIController')
var AWSS3 = require('./AWSS3')
const path = require('path');
let request = require("request-promise").defaults({ jar: true });

class PostingController {
    projectRefMap;
    folderRefMap;
    defectRefMap;
    tcIdRefMap;
    tcNameRefMap;
    tcStepRefMap;
    usRefMap = new Map();
    tsRefMap
    userRef;
    testcaseCollection = [];
    defectCollection = [];
    tcResultCollection = [];
    testsToBePosted = [];
    iterationRef;
  //  testSetRef;
    removeTagsCollection = [];
    addTagCollection = [];
    tsCollection = [];
    tagsToCreate = [];
    logs = [];
    trendsJSON; 
    updateTCR;
    testSetUpdateEntryCollection = [];
    /*
        Calls all the functions needed for posting all rally properties using th parsed data. At this location APIKey
        presence need not be needed as its already validated before calling this function
        as */
    async logResults(parsedResults, rallyProcessingObj, generateTrends, trendsObj) {
        try {
            await rallyProcessingObj.makeConnection();
            this.results = parsedResults;
            this.trendsJSON = parsedResults.trendsJSON;
            this.rallyProcessingObj = rallyProcessingObj;
            this.userRef = await rallyProcessingObj.getUserRef();
            let proxy = await rallyProcessingObj.getRallyProxy();
            let apikey = await rallyProcessingObj.getRallyUserApiKey();
            await this.createReferenceMaps(apikey,proxy);
            let postingResult;
            let createNewTestcasesFlag = await rallyProcessingObj.getCreateNewTestCases();
            if (createNewTestcasesFlag === true || createNewTestcasesFlag === 'true')
                postingResult = this.results;
            else
                postingResult = this.testsToBePosted;
            this.updateTCR = await rallyProcessingObj.getUpdateTCR();
            for (let testResultObj of postingResult)
                await this.createTestCaseEntries(testResultObj);
            let tcPostResult = await RallyBatchAPIController.postTestCases(this.testcaseCollection, apikey, proxy, "Successfully logged testcase data", "Failed Posting and updating testcases due to internal error");
            this.logs.push( tcPostResult.logs + "   " + new Date() );
            this.tcNameRefMap = tcPostResult.allTCRefs;
            for (let testResultObj of postingResult)
                await this.createAllTypesOfEntries(testResultObj);
            if( this.testSetUpdateEntryCollection.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.testSetUpdateEntryCollection, apikey, proxy, "Successfully updated testsets with testcases", "Failed updating testsets with testcases", "Adding Tags") + "   " +  new Date());
            else
                this.logs.push("No Testsets are updated");
            if( this.tcResultCollection.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.tcResultCollection, apikey, proxy, "Successfully posted testcase results" , "Internal Error Posting Testcase results to Rally.", "Testcase Results")  + "   " + new Date());
            else
                this.logs.push("No Testcase results added or updated");
            if( this.defectCollection.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.defectCollection, apikey, proxy, "Successfully linked defects with testcases", "Internal Error Posting Testcase results to Rally.", "Defects") + "   " +  new Date() );
            else
                this.logs.push("No Defects are linked with testcases");
            if( this.tagsToCreate.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.tagsToCreate, apikey, proxy, "Successfully created new tags", "Failed creating new tags","New Tags") + "   " +  new Date() );
            else
                this.logs.push("No new tags are created");
            if( this.removeTagsCollection.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.removeTagsCollection, apikey, proxy, "Successfully removed unspecified tags on testcases", "Failed removing unspecified tags on testcases", "Delete Tags") + "   " +  new Date() );
            else
                this.logs.push("No tags on testcases are removed");    
            if( this.addTagCollection.length > 0)
                this.logs.push(await RallyBatchAPIController.performBatchOperation(this.addTagCollection, apikey, proxy, "Successfully added specified tags to testcases", "Failed adding specified tags to testcases", "Adding Tags") + "   " +  new Date());
            else
                this.logs.push("No new tags are attached to testcases"); 
            let trendsPostingResult = { errors: false, msg: "Skipping Trends Posting.\n"};
            if ( generateTrends && (generateTrends === 'true' || generateTrends === true))
                trendsPostingResult = await this.mapandPostTrendsJSON(trendsObj);
            return { errors: trendsPostingResult.errors, postingError: false, trendsPostingError: trendsPostingResult.errors, message: this.logs.join("\n") + "\n"+ trendsPostingResult.msg }
        } catch (error) {
            return { errors: true, postingError: true, generateTrends: false, message: (this.logs.length> 0? this.logs.join("\n") : "" )+"Rally posting failed due to below error: \n" + error.message + "\nDue to rally posting failure, skipping trends posting.\n" }
        }
    }

    async mapandPostTrendsJSON(trendsObj) {
        let trendsJSONArr = [];
        for ( let trendRecord of this.results ) {
            let trendsJSON = {};
            let scenarioName = await trendRecord.getTestName();
            let testcaseId = await trendRecord.getTestCaseID();
            let folderDet = this.folderRefMap.get(await trendRecord.getTestFolderPath() + "~#~" + await trendRecord.getProjectName());
            let featureFileName = path.win32.basename(await trendRecord.getFeatureNamePath());
            featureFileName = featureFileName.substring(0, featureFileName.indexOf('.'));
            trendsJSON["featureName"] = await trendRecord.getFeatureName();
            trendsJSON["featureFileName"] = featureFileName;
            trendsJSON["RallyFolderID"] = folderDet? folderDet.FormattedID : null;
            trendsJSON["ScenarioName"] = scenarioName;
            let testcaseObj;
            if(testcaseId) {
                testcaseObj = this.tcIdRefMap.get(testcaseId);
            }
            else {
                testcaseObj = this.tcNameRefMap.get(scenarioName + ":" + folderDet._ref.slice(folderDet._ref.indexOf('/testfolder')));
            }
            trendsJSON["TestCaseID"] = testcaseObj? testcaseObj.FormattedID : null;
            trendsJSON["ScenarioStatus"] = await trendRecord.getTestStatus();
            trendsJSON["ExecutionTime"] = (new Date()).toISOString().slice(0, 19).replace("T", " ");
            trendsJSONArr.push(trendsJSON);
        }
        let json = JSON.stringify(trendsJSONArr, ',').replace('[', '').replace(new RegExp('},', 'g'), '}\n');
        let updatedjson = json.substring(0,json.lastIndexOf(']'));
        return AWSS3.postTrendsJson(updatedjson, await trendsObj.getFrameWorkFolder(),await trendsObj.getAwss3Proxy(), await trendsObj.getTrendsModule(), await trendsObj.getTrendsEnv(), await trendsObj.getTrendsJobname());
    }

    async createReferenceMaps(apikey,proxy) {

        this.projectRefMap = await this.createProjectRefMap();
        if (this.projectRefMap.get('errors'))
            throw new Error("Error While Fetching Project references in Rally.\nDetails : " + this.projectRefMap.get('message'));
        //
        // get references to existing folders and create if they do not already
        // exist

        this.folderRefMap = await this.createFolderRefMap();
        if (this.folderRefMap.get('errors'))
            throw new Error("Error While Fetching Folder references in Rally.\nDetails : " + this.folderRefMap.get('message'));
        // get references to existing defect
        this.defectRefMap = await this.createDefectRefMap();
        if (this.defectRefMap.get('errors'))
            throw new Error("Error While Fetching Defect references in Rally.\nDetails : " + this.defectRefMap.get('message'));
        // get references to test cases, creating new as needed
        let tcRefMap = await this.createTCRefMap();
        if (tcRefMap.get('errors'))
            throw new Error("Error While Fetching Testcase references in Rally.\nDetails : " + tcRefMap.get('message'));
        else {
            this.tcIdRefMap = tcRefMap.get("tcIdRefMap");
            this.tcNameRefMap = tcRefMap.get("tcNameRefMap");
            this.logs.push(tcRefMap.get("logs"));
        }
        // get references to User Stories
        this.usRefMap = await this.createUSRefMap();
        if (this.usRefMap.get('errors'))
            throw new Error("Error While Fetching Userstory references in Rally.\nDetails : " + this.usRefMap.get('message'));
        this.tsRefMap = await this.createTSRefMap(apikey,proxy);
        if (this.tsRefMap.get('errors'))
            throw new Error("Error While Fetching Testset references in Rally.\nDetails : " + this.tsRefMap.get('message'));
        this.tagsToCreate = await this.findandCreateTagEntries();
    }

    async createProjectRefMap() {
        const uniqueProjectNames = [...new Set(await Promise.all(this.results.map(async scenario => scenario.getProjectName())))].filter(function (e) {return e != null;});
        let projectRefMapLoc = new Map();
        let invalidProj = '';
        for (let projectName of uniqueProjectNames) {
            let reqedProjects  = await this.rallyProcessingObj.queryRallyForData('Project', 'Project', [{attrName : "Name", operator : "=", attrVal : projectName}], [] )
            if (reqedProjects.Results.length > 0) {
                projectRefMapLoc.set(projectName, reqedProjects.Results[0]._ref);
            } else {
                invalidProj += "Project " + projectName + " doesn't exist or there are no permissions for current user to access the project\n";
            }
        }
        if (!invalidProj)
            return projectRefMapLoc;
        else {
            let errorMap = new Map();
            errorMap.set("errors", true);
            errorMap.set("message", invalidProj);
            return errorMap;
        }
    }

    async createFolderRefMap() {
        let folderRefMapLoc = new Map();
        let folderAndProjCombi = await Promise.all(this.results.map(async scenario => {
            return {
                "folderPath": await scenario.getTestFolderPath(),
                "project": await scenario.getProjectName()
            }
        }));
        let iterate = true, folderLevel = 1;
        let parentfolderRef= new Map(); 
	    while ( iterate ) {
            let finalFolderListwithProjectNames = [], currItrParentFoldRefs= new Map();;
            const uniqueFolderProjectCombis = [...new Set(await Promise.all(folderAndProjCombi.map(async (combi, index) => {
                let indexAtnthRep = await this.getIndexAtNthOccurance(combi.folderPath,"/",folderLevel);
                if ( indexAtnthRep === combi.folderPath.length ) {
                    finalFolderListwithProjectNames.push(combi.folderPath + "~#~" + combi.project);
                    folderAndProjCombi.splice(index, 1);
                }
                return combi.folderPath.substring(0, indexAtnthRep) + "~#~" + combi.project
            })))].filter(function (e) {return e != null;});
            for ( let uniqueFolderProjectCombi of uniqueFolderProjectCombis) {
                let folderDet = await this.getOrCreateFolder(uniqueFolderProjectCombi, parentfolderRef);
                if ( finalFolderListwithProjectNames.includes(uniqueFolderProjectCombi) )
                    folderRefMapLoc.set(uniqueFolderProjectCombi, folderDet);
                else
                    currItrParentFoldRefs.set(uniqueFolderProjectCombi, folderDet._ref);
            }
            parentfolderRef = currItrParentFoldRefs;
            if(currItrParentFoldRefs.size === 0)
                iterate = false;
            folderLevel++;
        }
        return folderRefMapLoc;
    }

    async getOrCreateFolder(uniqueFolderPathProjCombi, parentProjRefs) {
        let folderInfo ={};
        let fpPnArr = uniqueFolderPathProjCombi.split("~#~");
        let folders = fpPnArr[0].split("/");
        let queryElem = [
            {
                attrName: "Name",
                operator: "=",
                attrVal: folders[folders.length-1]
            },
            {
                attrName: "Project",
                operator: "=",
                attrVal: this.projectRefMap.get(fpPnArr[1])
            }
        ]
        if(folders[folders.length-2]) {
            queryElem.push({
                attrName: "Parent",
                operator: "=",
                attrVal: parentProjRefs.get((folders.slice(0,[folders.length-1]).join("/")) + "~#~" + fpPnArr[1])
            })
        }
        let reqedFolder = await this.rallyProcessingObj.queryRallyForData("TestFolder", ["TestFolder","FormattedID"], queryElem, []);
        if (reqedFolder.Errors.length <= 0) {
            if (reqedFolder.Results.length > 0) {
                folderInfo._ref = reqedFolder.Results[0]._ref;
                folderInfo.FormattedID = reqedFolder.Results[0].FormattedID;
            } else {
                let data = {
                    "Name": folders[folders.length-1],
                    "Project": this.projectRefMap.get(fpPnArr[1])
                }
                if(folders[folders.length-2]) { data.Parent =  parentProjRefs.get((folders.slice(0,[folders.length-1]).join("/")) + "~#~" + fpPnArr[1]) }
                let createResult = await this.rallyProcessingObj.createRallyObject('TestFolder', data);
                if (createResult.Errors.length <= 0) {
                    this.logs.push("Created Test Folder " + folders[folders.length-1] +" with test folder id: " + createResult.Object.FormattedID + " in project: " + fpPnArr[1]);
                    folderInfo._ref = createResult.Object._ref;
                    folderInfo.FormattedID = createResult.Object.FormattedID;
                }
            }
        } else {
            throw new Error("Error while fetching TestFolder Details : " + reqedFolder.Errors.join("\n"));
        }
        return folderInfo;
    }

    async createDefectRefMap() {
        const uniqueDefectIds = [...new Set(await Promise.all(this.results.map(async scenario => scenario.getDefectId())))].filter((defect) => defect != null && defect != undefined);
        let defectRefMapLoc = new Map();
        let invalidDefects = '';
        for (let defectId of uniqueDefectIds) {
            let queryElem = [
                {
                    attrName: "FormattedID",
                    operator: "=",
                    attrVal: defectId
                }
            ]
            let reqedDefect = await this.rallyProcessingObj.queryRallyForData("Defect", "Defect", queryElem, []);
            if (reqedDefect.Results.length > 0) {
                defectRefMapLoc.set(defectId, reqedDefect.Results[0]._ref);
            } else {
                invalidDefects += defectId + "\n";
            }
        }
        if (!invalidDefects)
            return defectRefMapLoc;
        else {
            let errorMap = new Map();
            errorMap.set("errors", true);
            errorMap.set("message", "Invalid defects:\n" + invalidDefects);
            return errorMap;
        }
    }

    async createTCRefMap() {
        let tcIdRefMapLoc = new Map();
        let tcNameRefMapLoc = new Map();
        let invalidTC = [];
        let tcIdsQueries = [], tcNameQueries=[];
        let reqedTestcaseByid, reqedTCbyName;
        let tcIdDet = {}, tcNameDet = {};
        let repeatedScenarios = [];
        let logs ="";
        for (let scenario of this.results) {
            let tcId = await scenario.getTestCaseID();
            let projectName = await scenario.getProjectName();
            let folderPath = await scenario.getTestFolderPath();
            let tcName = await scenario.getTestName();
            let folderRef = this.folderRefMap.get(folderPath + "~#~" + projectName)._ref;
            if (tcId) {
                if (tcIdDet.hasOwnProperty(tcId)) {
                    //same named scenario found. removing the first occured scenario.
                    repeatedScenarios.push(tcIdDet[tcId].scenario)
                    tcIdDet[tcId].scenario = scenario;
                } else {
                    tcIdDet[tcId] = {
                        scenario : scenario,
                        folderRef : folderRef,
                        scenarioName : tcName,
                        projectName: projectName
                    }
                }
            } else {
                if (tcNameDet.hasOwnProperty(tcName + "~#~:" + folderRef)) {
                    //same named scenario found. removing the first occured scenario.
                    repeatedScenarios.push(tcNameDet[tcName + "~#~:" + folderRef].scenario)
                    tcNameDet[tcName + "~#~:" + folderRef].scenario = scenario;
                }
                else {
                    tcNameDet[tcName + "~#~:" + folderRef] = {
                        scenario : scenario,
                        folderRef : folderRef,
                        projectName: projectName
                    }
                }
            }
        }
        if( repeatedScenarios.length > 0) { logs += "\n----------------------------------------\nFound scenarios with same name or same testcaseids,\nPosting Last occurance,\nNames of scenarios skipped: \n"}
        for ( let scenario of repeatedScenarios) {logs += await scenario.getTestName() + '\n';this.results.splice(this.results.indexOf(scenario),1);}
        const noOfTcbyIdToGetPerItr = 500;
        const noOfTcbyNameToGetPerItr = 25;
        let noOfIterToGetAllTcbyId, noOfIterToGetAllTcbyName ;
        tcIdsQueries = Object.keys(tcIdDet).map((tcId) => {return {
            attrName: "FormattedID",
            operator: "=",
            attrVal: tcId
        }});
        let nameNoMatchTCIds = [];
        if(tcIdsQueries.length > noOfTcbyIdToGetPerItr)  { noOfIterToGetAllTcbyId = (tcIdsQueries.length % noOfTcbyIdToGetPerItr) === 0 ? tcIdsQueries.length / noOfTcbyIdToGetPerItr :  Math.floor(tcIdsQueries.length / noOfTcbyIdToGetPerItr) + 1  }
        else noOfIterToGetAllTcbyId = 1;
        
        for (let iterationCount = 0 ; iterationCount < noOfIterToGetAllTcbyId ; iterationCount++ ) {
            let tcDetToFetch = tcIdsQueries.splice( 0, noOfTcbyIdToGetPerItr );
            if ( tcDetToFetch.length > 0 ) {
                reqedTestcaseByid = await this.rallyProcessingObj.queryRallyForData("Testcase", ["Name","Testcase", "FormattedID", "LastResult", "Tags", "Testset"], [], tcDetToFetch);
                //once invalidTc has value tcIdRefMapLoc values need not be considered. so not adding non invalidTc in if block to tcIdRefMapLoc
                if ( reqedTestcaseByid.Results.length  < tcDetToFetch.length ) {
                    let resultedTcIds = reqedTestcaseByid.Results.map((tcRes) => tcRes.FormattedID);
                    invalidTC.push(...tcDetToFetch.filter((givenTc) => !(resultedTcIds.includes(givenTc))).map((tcId) => "Invalid testcaseId: " + tcId + " in scenario : '" + tcIdDet[tcId].scenarioName + "' of feature : '" + tcIdDet[tcId].featurePath + "'" ));
                } else {
                    this.testsToBePosted.push(...reqedTestcaseByid.Results.map((tcRes) => tcIdDet[tcRes.FormattedID].scenario));
                    let mapsInfoToSet = reqedTestcaseByid.Results.map((tcRes) => {if(tcRes.Name !== tcIdDet[tcRes.FormattedID].scenarioName ) {nameNoMatchTCIds.push(tcRes.FormattedID);} return [tcRes.FormattedID , {_ref: tcRes._ref, lastResult : tcRes.LastResult, tags: tcRes.Tags, FormattedID : tcRes.FormattedID} ]});
                    tcIdRefMapLoc = new Map([...tcIdRefMapLoc, ...mapsInfoToSet]);
                }
            } 
        }
        tcNameQueries = Object.keys(tcNameDet).map((tcNameFolderRefCombi) => {
            let tcFolderArr = tcNameFolderRefCombi.split("~#~:");

            return [{
            attrName: "Name",
            operator: "=",
            attrVal: tcFolderArr[0]
        },
        {
            attrName: "Project",
            operator: "=",
            attrVal: this.projectRefMap.get(tcNameDet[tcNameFolderRefCombi].projectName)
        },
        {
            attrName: "TestFolder",
            operator: "=",
            attrVal: tcFolderArr[1]
        }]})
        if(tcNameQueries.length > noOfTcbyNameToGetPerItr)  { noOfIterToGetAllTcbyName = (tcNameQueries.length % noOfTcbyNameToGetPerItr) === 0 ? tcNameQueries.length / noOfTcbyNameToGetPerItr :  Math.floor(tcNameQueries.length / noOfTcbyNameToGetPerItr) + 1  }
        else noOfIterToGetAllTcbyName = 1;
        let repeatedByIdAndName= [];
        for (let iterationCount = 0 ; iterationCount < noOfIterToGetAllTcbyName ; iterationCount++ ) {
            let tcDetToFetch = tcNameQueries.splice( 0, noOfTcbyNameToGetPerItr );
            if ( tcDetToFetch.length > 0 ) {
                reqedTCbyName = await this.rallyProcessingObj.complexQueryRallyForData("Testcase", ["Testcase", "Name", "FormattedID","LastResult", "Tags", "TestFolder", "TestSets"], tcDetToFetch, []);
                //even though the tc with name doesn't exists it is not considered as invalid. we have to create tc with name based on createnewTc header value
                this.testsToBePosted.push(...new Set(reqedTCbyName.Results.map((tcRes) => tcNameDet[tcRes.Name + "~#~:" + tcRes.TestFolder._ref].scenario)));
                tcNameRefMapLoc = new Map([...tcNameRefMapLoc,...(await Promise.all(reqedTCbyName.Results.map(async (tcRes) => {
                    if(nameNoMatchTCIds.includes(tcRes.FormattedID)) { repeatedByIdAndName.push(tcNameDet[tcRes.Name + "~#~:" + tcRes.TestFolder._ref].scenario)};
                    return [tcRes.Name + ":" + tcRes.TestFolder._ref.slice(tcRes.TestFolder._ref.indexOf('/testfolder')) , {_ref: tcRes._ref, lastResult : tcRes.LastResult, tags: tcRes.Tags, FormattedID : tcRes.FormattedID} ]
                })))]);
            }                       
        }   
        if( repeatedByIdAndName.length > 0) { logs += "\n----------------------------------------\nFound scenarios with testcaseId with different name and scenario updating same testcase, \nSkipping repeated testcase with name,\nNames of scenarios skipped: \n"}
        for ( let scenario of repeatedByIdAndName) {
            logs += await scenario.getTestName() + '\n';
            this.results.splice(this.results.indexOf(scenario),1);
            this.testsToBePosted.splice(this.testsToBePosted.indexOf(scenario), 1);
        }

        if ( !invalidTC.length > 0)
            return new Map([['tcIdRefMap', tcIdRefMapLoc],['tcNameRefMap',tcNameRefMapLoc], ['logs', logs]]);
        else {
            let errorMap = new Map();
            errorMap.set("errors", true);
            errorMap.set("message", "invalid TestCases:" + invalidTC.join("\n"));
            return errorMap;
        }
    }

    async createUSRefMap() {
        const uniqueUserStories = [...new Set(await Promise.all(this.results.map(async scenario => scenario.getUserStory())))].filter(function (e) {return e != null;});
        let USRefMapLoc = new Map();
        let invalidUS = '';
        for (let userStory of uniqueUserStories) {
            let queryElem = [
                {
                    attrName: "FormattedID",
                    operator: "=",
                    attrVal: userStory
                }
            ]
            let requestedUSs = await this.rallyProcessingObj.queryRallyForData("hierarchicalrequirement", "FormattedId", queryElem, []);
            if (requestedUSs.Results.length > 0) {
                USRefMapLoc.set(userStory, requestedUSs.Results[0]._ref);
            } else {
                invalidUS += userStory + "\n" ;
            }
        }
        if (!invalidUS)
            return USRefMapLoc;
        else {
            let errorMap = new Map();
            errorMap.set("errors", true);
            errorMap.set("message", "Invalid userstories ids:\n" + invalidUS);
            return errorMap;
        }
    }

    async findandCreateTagEntries() {
        let allTagsOfTcs = await Promise.all(this.results.map(async scenario => { 
            let tags = await scenario.getTags();
            if ( tags ) {
                return tags.split('|');
            }  else return '';
        }));
        const uniqueTagNames = [...new Set(allTagsOfTcs.flat(1).filter(tag => tag != ''))];
        let tagsToCreate = [];
        let tagQuery = uniqueTagNames.map((tag) => {return {
            attrName: "Name",
            operator: "=",
            attrVal: tag
        }})
        let noOfIterToGetAllTagsbyName, noOfTagbyNameToGetPerItr = 100;
        if(tagQuery.length > noOfTagbyNameToGetPerItr)  { noOfIterToGetAllTagsbyName = (tagQuery.length % noOfTagbyNameToGetPerItr) === 0 ? tagQuery.length / noOfTagbyNameToGetPerItr :  Math.floor(tagQuery.length / noOfTagbyNameToGetPerItr) + 1  }
        else noOfIterToGetAllTagsbyName = 1;
        
        for (let iterationCount = 0 ; iterationCount < noOfIterToGetAllTagsbyName ; iterationCount++ ) {
            let tagDetToFetch = tagQuery.splice( 0, noOfTagbyNameToGetPerItr);
            let tagsToFetch = uniqueTagNames.splice( 0, noOfTagbyNameToGetPerItr);

            if ( tagDetToFetch.length > 0 ) {
                let reqedTagsbyName = await this.rallyProcessingObj.queryRallyForData("Tag", ["Name"], [], tagDetToFetch);
                let reqedTagNames = reqedTagsbyName.Results.map((tag) => tag.Name);
                tagsToCreate.push(...await Promise.all(tagsToFetch.filter(tagToFetch => !reqedTagNames.includes(tagToFetch))
                .map(async (tagName) => await this.rallyProcessingObj.createEntry("tag", '/tag/create', 'POST', {Name : tagName})
                )  ) ) 
            }                       
        } 
        return tagsToCreate;

    }

    async createTSRefMap(apikey,proxy) {
        const uniqueTestSets = [...new Set(await Promise.all(this.results.map(async scenario => scenario.getTestSetId())))].filter(function (e) {return e != null;});
        let TSRefMapLoc = new Map();
        let invalidTS = '';
        for (let testset of uniqueTestSets) {
            let queryElem = [
                {
                    attrName: "FormattedID",
                    operator: "=",
                    attrVal: testset
                }
            ]
            let requestedTSs = await this.rallyProcessingObj.queryRallyForData("Testset", ["Testset","TestCases"], queryElem, []);
            if (requestedTSs.Results.length > 0) {
                let options = {
                    method: "GET",
                    url: requestedTSs.Results[0].TestCases._ref,
                    headers: {zsessionid: apikey},
                    proxy: proxy
                };
                let response = JSON.parse(await request(options));
                TSRefMapLoc.set(testset, {_ref: requestedTSs.Results[0]._ref, testcases:response.QueryResult.Results.map((tcDet) => {return {_ref: tcDet._ref}})});

            } else {
                invalidTS += testset + "\n" ;
            }
        }
        if (!invalidTS)
            return TSRefMapLoc;
        else {
            let errorMap = new Map();
            errorMap.set("errors", true);
            errorMap.set("message", "Invalid testset ids:\n" + invalidTS);
            return errorMap;
        }
    }
 
    // Create Entries for Testcases 
    async createTestCaseEntries(testResultScenario) {
        let testcaseDet = {};
        testcaseDet.Owner = this.userRef;
        let projectName =  await testResultScenario.getProjectName();
        testcaseDet.Project = this.projectRefMap.get(projectName);
        let testFolderRef = this.folderRefMap.get(await testResultScenario.getTestFolderPath() + "~#~" + projectName)._ref;
        testcaseDet.TestFolder = testFolderRef;

        let method = await testResultScenario.getExecutionType();
        if (method)
            testcaseDet.Method = method;
        let desc = await testResultScenario.getDescription();
        if (desc)
            testcaseDet.Description = desc;
        let type = await testResultScenario.getTestType();
        if (type)
            testcaseDet.Type = type;
        let risk = await testResultScenario.getRisk();
        if (risk)
            testcaseDet.Risk = risk;
        let priority = await testResultScenario.getPriority();
        if (priority)
            testcaseDet.Priority = priority;
        let objective = await testResultScenario.getObjective();
        if (objective)
            testcaseDet.Objective = objective;
        let notes = await testResultScenario.getNotes();
        if (notes)
            testcaseDet.Notes = notes;
        let userStory = this.usRefMap.get(await testResultScenario.getUserStory());
        if (userStory)
            testcaseDet.WorkProduct = userStory;
        let preconditions = await testResultScenario.getPreConditions();
        if (preconditions)
            testcaseDet.PreConditions = preconditions;
        let postconditions = await testResultScenario.getPostConditions();
        if (postconditions)
            testcaseDet.PostConditions = postconditions;
        let testcaseId = await testResultScenario.getTestCaseID();
        let tcKeyRefObj, testcaseName = await testResultScenario.getTestName();

        if(testcaseId) {
            tcKeyRefObj = this.tcIdRefMap.get(testcaseId);
        }
        else {
            tcKeyRefObj = this.tcNameRefMap.get(testcaseName + ":" + testFolderRef.slice(testFolderRef.indexOf('/testfolder')));
        }
        let entrymethod, path;
        if (tcKeyRefObj) {
            entrymethod = 'PUT';
            path = tcKeyRefObj._ref.slice(tcKeyRefObj._ref.indexOf('/testcase'));
        }
        else {
            path = "/testcase/create";
            this.logs.push("Creating testcase with name: " + testcaseName)
            entrymethod = 'POST';
        }
        //create with scenario name if tc doesn't exisits if not update tc with name given ( if tcId is specified)
        testcaseDet.Name = testcaseName;
        let testcaseEntry = await this.rallyProcessingObj.createEntry("testcase", path, entrymethod, testcaseDet)
        this.testcaseCollection.push(testcaseEntry);
    
    }

    // Create Entries for Defects and testcaseresults and testsets
    async createAllTypesOfEntries(testResultScenario) {
        let tcResultDet = {  };
        let tcId = await testResultScenario.getTestCaseID();
        let tcName = await testResultScenario.getTestName();
        let projectName = await testResultScenario.getProjectName();
        let fodlerPath = await testResultScenario.getTestFolderPath();
        let tcKeyRefObj ;
        let testFolderRef = this.folderRefMap.get(fodlerPath + "~#~" + projectName)._ref;
        if ( tcId ) {
            tcKeyRefObj = this.tcIdRefMap.get(tcId);
        } else {
            let tcRefkeyName = tcName + ":" + testFolderRef.slice(testFolderRef.indexOf('/testfolder'));
            tcKeyRefObj = this.tcNameRefMap.get(tcRefkeyName);
        }
        //Its mandatory to have tcRef at this locations, as all the refs are mapped right after 
        let tcRef = tcKeyRefObj._ref
        let tagsStr = await testResultScenario.getTags();
        let tags = tagsStr ? [...new Set(tagsStr.split('|').filter(tag => tag))] : [];
        if (tags.length > 0 ) {
            let removalIndecies =[];
            if(tcKeyRefObj.tags && tcKeyRefObj.tags._tagsNameArray.length > 0) {
                tcKeyRefObj.tags._tagsNameArray.map((exisitingTag, index) => {
                    let indexOfExisitingTag = tags.indexOf(exisitingTag.Name);
                    if(indexOfExisitingTag > -1) { 
                        tags.splice(indexOfExisitingTag, 1);
                        removalIndecies.push(index);
                    }   
                })
                removalIndecies.sort().reverse();
                for(let index of removalIndecies) { tcKeyRefObj.tags._tagsNameArray.splice(index, 1); }
                if(tcKeyRefObj.tags._tagsNameArray.length > 0)
                    this.removeTagsCollection.push(await this.rallyProcessingObj.createEntry("CollectionItems", tcRef.slice(tcRef.indexOf('/testcase'))  + "/tags/remove", "POST",tcKeyRefObj.tags._tagsNameArray.map((tag) => {
                        return {
                            "_ref" : tag._ref
                        }
            }) ));
            }
            if( tags.length > 0) {
                this.addTagCollection.push(await this.rallyProcessingObj.createEntry("CollectionItems", tcRef.slice(tcRef.indexOf('/testcase'))  + "/tags/add", "POST", tags.map((tag) => {
                    return {
                        "Name" : tag
                    }
                }) ));
            }
        }
        tcResultDet.TestCase = tcRef;
        tcResultDet.Tester = this.userRef;
        let build = await testResultScenario.getBuild();
        tcResultDet.Build = build === null ? 1 : build;
        tcResultDet.Verdict = await testResultScenario.getTestStatus();
        tcResultDet.Duration = await testResultScenario.getDurationMS();
        tcResultDet.Date = new Date();
        let notes = await testResultScenario.getNotes();
        let status = await testResultScenario.getTestStatus();
        let stackTrace = await testResultScenario.getStackTrace();
        if (status && status === 'Fail')
            notes += stackTrace === null ? '' : stackTrace;
        if (notes)
            tcResultDet.Notes = notes;
        let tsId = await testResultScenario.getTestSetId();
        if (tsId) {
            let tsRefObj = this.tsRefMap.get(tsId);
            let tsRef = tsRefObj._ref;
            let testcasesOfTS = tsRefObj.testcases;
            this.tsRefMap.set(tsId,{_ref:tsRef, TestCases: null});
            tcResultDet.TestSet = {};
            tcResultDet.TestSet._ref = tsRef;
            let updatedTCOfTS = [{_ref : tcRef}];
            if (testcasesOfTS){
                updatedTCOfTS = [...updatedTCOfTS, ...testcasesOfTS];
            }
            this.testSetUpdateEntryCollection = await this.rallyProcessingObj.findEntryByPathAndAddToBody(this.testSetUpdateEntryCollection, tsRef.slice(tsRef.indexOf("/testset")), updatedTCOfTS, "testset","TestCases");
        }
        let tcrPath = "/testcaseresult/create";
        if(this.updateTCR) {
            if ( tcKeyRefObj.lastResult )
                tcrPath = tcKeyRefObj.lastResult._ref.slice(tcKeyRefObj.lastResult._ref.indexOf('/testcaseresult')) ;
        }
        let tcResultEntry = await this.rallyProcessingObj.createEntry("testcaseresult", tcrPath, "POST", tcResultDet);
        this.tcResultCollection.push(tcResultEntry);
        let defectId = await testResultScenario.getDefectId();
        if (defectId) {
            let defRef = this.defectRefMap.get(defectId);
            let defectEntry = await this.rallyProcessingObj.createEntry("defect", defRef.slice(defRef.indexOf('/defect')), "PUT", { "TestCase": tcRef, "Requirement" : null });
            this.defectCollection.push(defectEntry);
        }
    }

    async getIndexAtNthOccurance(string, subString, index) {
        let splittedStr = string.split(subString, index);
        let joinedStr = splittedStr.join(subString);
        return joinedStr.length;
    }
}

module.exports = PostingController;