let rally = require('rally');
let queryUtils = rally.util.query;
let RallyProcessing = class  {

	#mRallyUserName = null;
	#mRallyUserApiKey = null;
	#mRallyProxy = null;
	#mCreateNewTestCases = null;
    #mUpdateTCR = null;
	restApi;
	invalidDataCollection = null;

	constructor() {
		this.invalidDataCollection = [];
	}

	async verifyInvalidAttributeValues(attributeName, attributeValue) {
		if (attributeValue == null)
			this.invalidDataCollection.push("Missing '" + attributeName + "' header.");
		if(((typeof attributeValue === 'string' || attributeValue instanceof String) && ( attributeValue.length <= 0 || attributeValue.trim().length <= 0
			|| attributeValue.length > 32768))) {
			this.invalidDataCollection.push("Invalid '" + attributeName + "' header value.");
		}
	}

	async getRallyUserName() {
		return this.#mRallyUserName;
	}

	async setRallyUserName(mRallyUserName) {
		if (mRallyUserName)
			this.#mRallyUserName = mRallyUserName;
		else 
			this.#mRallyUserName = "shared.rallyintegrationuser@ge.com";

		this.verifyInvalidAttributeValues("RallyUserName", this.#mRallyUserName);
	}


	async getRallyUserApiKey() {
		return this.#mRallyUserApiKey;
	}

	async setRallyUserApiKey(mRallyUserApiKey, headerName) {
		this.#mRallyUserApiKey = mRallyUserApiKey;
		this.verifyInvalidAttributeValues(headerName, this.#mRallyUserApiKey);
	}

	async getRallyProxy() {
		return this.#mRallyProxy;
	}

	async setRallyProxy(mRallyProxy) {
		if (mRallyProxy) 
			this.#mRallyProxy = mRallyProxy;
		else 
			this.#mRallyProxy = "http://502635365:RIU.8272.ge@PITC-Zscaler-Americas-Alpharetta3pr.proxy.corporate.ge.com:80"
		this.verifyInvalidAttributeValues("RallyProxy", this.#mRallyProxy);
	}

	async getCreateNewTestCases() {
		return this.#mCreateNewTestCases;
	}

	async setCreateNewTestCases(mCreateNewTestCases) {
		this.#mCreateNewTestCases = mCreateNewTestCases;
	}

    async getUpdateTCR() {
		return this.#mUpdateTCR;
	}

    async setUpdateTCR(mUpdateTCR) {
		this.#mUpdateTCR = mUpdateTCR && (mUpdateTCR === 'true' || mUpdateTCR === true);
	}

	async makeConnection() {
        this.restApi = await rally({
            apiKey: await this.getRallyUserApiKey(),
            requestOptions: {
                headers: {
                    'X-RallyIntegrationName': 'Rally Posting API',
                    'X-RallyIntegrationVendor': 'GE Corporate',
                    'X-RallyIntegrationVersion': '1.0'
                }, 
				proxy: await this.getRallyProxy() ,
				jar: true
            }
        });
    }
	
	async getUserRef() {
        try {
            let qr = queryUtils.where("Name", "=", (await this.getRallyUserName()).toLowerCase());
            let reqedUser = await this.restApi.query({
                type: 'User', //the type to query
                fetch: ['User'], //the fields to retrieve
                query: qr
            });
            if (reqedUser.Results.length > 0) {
                return reqedUser.Results[0]._ref;
            } else {
                throw new Error("invalid user")
            }
        } catch (e) {
            if (e.message === 'invalid user')
                throw new Error("Rally posting failed. Make sure you have provided valid rally user account credentials.");
            else
                throw new Error('Connection issue. Please try after some time')
        }
    }

	async queryRallyForData(type, fetchParamsArr, andQueryElements, orQueryElements) {
		let qr;
		if ( andQueryElements.length > 0 ) {
			qr = queryUtils.where(andQueryElements[0].attrName, andQueryElements[0].operator, andQueryElements[0].attrVal);
			andQueryElements.splice(0,1);
			for(let queryInfo of andQueryElements) 
				qr = qr.and(queryInfo.attrName, queryInfo.operator, queryInfo.attrVal);
		} else {
			if ( orQueryElements.length > 0 ) {
				qr = queryUtils.where(orQueryElements[0].attrName, orQueryElements[0].operator, orQueryElements[0].attrVal);
				orQueryElements.splice(0,1);
			}
		}
		for ( let queryInfo of orQueryElements ) 
			qr = qr.or(queryInfo.attrName, queryInfo.operator, queryInfo.attrVal)
		return this.restApi.query({
			type: type, //the type to query
			start: 1, //the 1-based start index, defaults to 1
			pageSize: 200, //the page size (1-200, defaults to 200)
			limit: Infinity, //the maximum number of results to return- enables auto paging
			fetch: fetchParamsArr, //the fields to retrieve
			query: qr
		});

	}
	// AND of ORs [( A or B) and ( C  or D)....] and OR od ANDs [( A or B) and ( C or D)...] are joined using or. 
	async complexQueryRallyForData( type, fetchParamsArr, orOfAndsQueryElements, andOfOrsQueryElements ) {
		let qr;
		if ( orOfAndsQueryElements.length > 0 ) {
			orOfAndsQueryElements.forEach((andofOrQueryInfo, index) => {			
				let andQr = queryUtils.where(andofOrQueryInfo[0].attrName, andofOrQueryInfo[0].operator, andofOrQueryInfo[0].attrVal);
				andofOrQueryInfo.splice(0,1);
				for ( let orQueryInfo of andofOrQueryInfo)
					andQr = andQr.and(orQueryInfo.attrName, orQueryInfo.operator, orQueryInfo.attrVal);
				if ( index == 0) 
					qr = andQr;
				else
					qr = qr.or(andQr);

			});
		} else {
			if ( andOfOrsQueryElements.length > 0 ) {
				qr = queryUtils.where(andOfOrsQueryElements[0][0].attrName, andOfOrsQueryElements[0][0].operator, andOfOrsQueryElements[0][0].attrVal);
				andOfOrsQueryElements[0].splice(0,1);
			}
		}
		for ( let queryInfo of andOfOrsQueryElements ) 
			qr = qr.or(queryInfo.attrName, queryInfo.operator, queryInfo.attrVal)
		return this.restApi.query({
			type: type, //the type to query
			start: 1, //the 1-based start index, defaults to 1
			pageSize: 200, //the page size (1-200, defaults to 200)
			limit: Infinity, //the maximum number of results to return- enables auto paging
			fetch: fetchParamsArr, //the fields to retrieve
			query: qr
		});
	}

	async createRallyObject(type, data, fetchProjecties, scope) {

        return this.restApi.create({
            type: type, //the type to create
            data: data,
            fetch: fetchProjecties ? fetchProjecties : ['FormattedID'],  //the fields to be returned on the created object
            scope: scope,
            requestOptions: {} //optional additional options to pass through to request
        })
    } 

	async updateRallyObject(ref, data, properties) {

        return this.restApi.update({
            ref: ref, //the ref to update
            data: data,
            fetch: ['FormattedID'],  //the fields to be returned on the update object
            scope: properties,
            requestOptions: {} //optional additional options to pass through to request
        })
    } 

	async findEntryByPathAndAddToBody(allEntries, path,bodyToAdd, type, bodyProperty) {
			let entryToBeUpdated = allEntries.find((entry) => { if(entry.Entry.Path === path) {entry.Entry.Body[type][bodyProperty].push(...bodyToAdd); return entry;}});
		if(!entryToBeUpdated) {
			allEntries.push(await this.createEntry(type, path, "PUT", {[bodyProperty]: bodyToAdd} ));
		}
		return allEntries
	}

	async createEntry(type, path, method, body ) {
		let bodyObj  = {}
		if ( body ) {
			bodyObj = {
				[type]: body
			}
		}
		return {
            "Entry": {
                "Path": path,
                "Method": method,
                "Body": bodyObj
            }
        }
	}
	
}
module.exports = RallyProcessing;