let request = require("request-promise").defaults({ jar: true });

class RallyBatchAPIController {
    constructor() {
        this.baseUrl = "https://rally1.rallydev.com/slm/webservice/v2.0";
    }

    async postTestCases(testCaseCollection, apikey, proxy, successLog, errorLog ) {
        const numOfTCToPostPerBatch = 500; let errorsFlag = false, log = errorLog;
        let numOfIterationToPostAllTC 
        if(testCaseCollection.length > numOfTCToPostPerBatch)  { numOfIterationToPostAllTC = (testCaseCollection.length % numOfTCToPostPerBatch) === 0 ? testCaseCollection.length / numOfTCToPostPerBatch :  Math.floor(testCaseCollection.length / numOfTCToPostPerBatch) + 1  }
        else numOfIterationToPostAllTC = 1;
        let refMaps = new Map();
        for (let iterationCount = 0 ; iterationCount < numOfIterationToPostAllTC ; iterationCount++ ) {
            let entrToPostInCurIter = {"Batch" : []}
            entrToPostInCurIter.Batch = testCaseCollection.splice(0, numOfTCToPostPerBatch) ;
            let options = {
                method: "POST",
                url: this.baseUrl + "/batch?shared=true&fetch=true",
                headers: {zsessionid: apikey},
                body: JSON.stringify(entrToPostInCurIter),
                proxy: proxy
            };
            let response = JSON.parse(await request(options));
            
            if (response.BatchResult && response.BatchResult.Errors.length <= 0 && (response.OperationResult == null || (response.OperationResult && response.OperationResult.Errors <=0))  ) {
                refMaps = new Map([...refMaps,...(await this.mapTestCasesWithRef(response.BatchResult.Results))]);
            }
            else {
                let errorInfo = await this.printBulkUpdateErrors(response, "Testcases");
                if(errorInfo.errors) { errorsFlag = true; log += errorInfo.info; }
            }
        }
        if( !errorsFlag )  log = successLog;
        else throw new Error(log);
        return { logs: log, allTCRefs: refMaps};
    }

    async mapTestCasesWithRef(respResults) {
        let tcRefMap = new Map();
        for(let result of respResults) {
            //let scenario = Promise.all(results.find(async (scenario) => {if(await scenario.)}));
            tcRefMap.set(result.Object.Name + ":" + result.Object.TestFolder._ref.slice(result.Object.TestFolder._ref.indexOf('/testfolder')), {_ref: result.Object._ref, lastResult : result.Object.LastResult, tags : result.Object.Tags, FormattedID: result.Object.FormattedID } )
        }
        return tcRefMap;
    }   

    async performBatchOperation(entryCollection, apikey, proxy, successLog, errorLog, collectionType ) {
        const entriesPerBatch = 500;
        let numOfItrToPostAllEntries; 
        let errorsFlag = false, log = errorLog;
        if(entryCollection.length > entriesPerBatch)  { numOfItrToPostAllEntries = (entryCollection.length % entriesPerBatch) === 0 ? entryCollection.length / entriesPerBatch :  Math.floor(entryCollection.length / entriesPerBatch) + 1  }
        else numOfItrToPostAllEntries = 1;
        for (let iterationCount = 0 ; iterationCount < numOfItrToPostAllEntries ; iterationCount++ ) {
            let entrToPostInCurIter = {"Batch" : []}
            entrToPostInCurIter.Batch = entryCollection.splice(0, entriesPerBatch) ;
            let options = {
                method: "POST",
                url: this.baseUrl + "/batch?shared=true&fetch=true",
                headers: {zsessionid: apikey},
                body: JSON.stringify(entrToPostInCurIter),
                proxy: proxy
            };
            let response = JSON.parse(await request(options));
            let errorInfo = await this.printBulkUpdateErrors(response, collectionType);
            if(errorInfo.errors) { errorsFlag = true; log += errorInfo.info; }
        }
        if( !errorsFlag) log = successLog;
        else throw new Error(log);

        return log;
    }

    async printBulkUpdateErrors(response, collectionType) {
        let errors = []
        if(response.OperationResult != null) {
            if(response.OperationResult.Errors && response.OperationResult.Errors.length > 0)
                errors.push(response.OperationResult.Errors);
        }
        if(response.BatchResult != null){
            if(response.BatchResult.Errors && response.BatchResult.Errors.length > 0 )
                errors.push(response.BatchResult.Errors);
        }
        if ( errors.length > 0 ) {
            let info = "Internal Rally Error while posting "+ collectionType+" : " + errors.join("\n");
            return {errors: true, info : info} ;
        }
        return false;
    }
}
module.exports = new RallyBatchAPIController();