var fs = require("fs");
var path = require('path');
var filePath = process.cwd();
var tcrefsFile = path.resolve(filePath, 'mock_data/tcref.json');
var RallyBatchAPIController = require('./RallyBatchAPIController.js')
var defRallyproxy = "http://502635365:RIU.8272.ge@PITC-Zscaler-Americas-Alpharetta3pr.proxy.corporate.ge.com:80";
let rally = require('rally');
let queryUtils = rally.util.query;
class RallyCleanup {
    
    async cleanupByIds(tcIds, headerConfigProps) {
        try {
            console.debug("cleanupByIds");
            let tcDelEntries ;
                
            let rallyProxy = headerConfigProps.rallyproxy;
            let apikey = headerConfigProps.apikey;
            let createDelEntriesAPI = await rally({
                apiKey: apikey,
                requestOptions: {
                    headers: {
                        'X-RallyIntegrationName': 'Rally Posting API',
                        'X-RallyIntegrationVendor': 'GE Corporate',
                        'X-RallyIntegrationVersion': '1.0'
                    }, proxy: rallyProxy ? rallyProxy : defRallyproxy, jar: true
                }
            });
            tcDelEntries = await this.createDeleteEntriesByIds(tcIds, createDelEntriesAPI);

        //    let refs= fs.readFileSync(tcrefsFile, 'utf8');            
        //    await RallyBatchAPIController.performBatchOperation(JSON.parse(refs), headerConfigProps.apikey, headerConfigProps["rallyProxy"] ? headerConfigProps["rallyProxy"] : defRallyproxy, "Sucessfully deleted", "Error while deleting");            
          
        } catch(ex) {
            console.debug("Error while cleaning up stuff" + ex.message);
            if(ex.message.includes("524")) {
                await this.cleanupByIds(tcIds, headerConfigProps)
            }
        }
    }

    async createDeleteEntriesByIds(tcIds, createDelEntriesAPI) {
        try {
            let tcIdsCollection = tcIds.split(",");
            let tcRefEntries = []
            const noOfTcToGetPerBatch = 500;
            let noOfIterToGetAllTc;
            let notFetchTcIds = [];
            if (tcIdsCollection.length > noOfTcToGetPerBatch) { noOfIterToGetAllTc = (tcIdsCollection.length % noOfTcToGetPerBatch) === 0 ? tcIdsCollection.length / noOfTcToGetPerBatch : Math.floor(tcIdsCollection.length / noOfTcToGetPerBatch) + 1 }
            else noOfIterToGetAllTc = 1;
            for (let iterationCount = 0; iterationCount < noOfIterToGetAllTc; iterationCount++) {
                let tcIdsToFetchPerItr = tcIdsCollection.splice(0, noOfTcToGetPerBatch) ;

                let qr = queryUtils.where("FormattedID", "in", tcIdsToFetchPerItr)
                let reqedTestcase = await createDelEntriesAPI.query({
                    type: 'Testcase', //the type to query
                    start: 1, //the 1-based start index, defaults to 1
                    pageSize: 200, //the page size (1-200, defaults to 200)
                    limit: Infinity, //the maximum number of results to return- enables auto paging
                    fetch: ['FormattedId'], //the fields to retrieve
                    scope: {},
                    query: qr
                });
                if (reqedTestcase.Errors.length <= 0) {
                    let fetchedIds = reqedTestcase.Results.map(res => res.FormattedID);
                    notFetchTcIds.push(...(tcIdsToFetchPerItr.filter(value => !(fetchedIds.includes(value)))));
                    tcRefEntries.push(...(reqedTestcase.Results.map((tcDet) => {
                        return {
                            "Entry": {
                                "Path": tcDet._ref.slice(tcDet._ref.indexOf('/testcase')),
                                "Method": "DELETE"
                            }
                        }
                    })))
                }
            }
            if (notFetchTcIds.length > 0)
                fs.writeFileSync(tcrefsFile,JSON.stringify(notFetchTcIds), 'utf8');      
            return tcRefEntries;
        } catch (ex) {
            console.error("Error Occured while fetching tc details : " + ex.message)
        }
    }

    async getAllTCWithNoFolderAndTcNameContains(tcNamePortion) {
        
    }

    async cleanupByProjectName(projectName) {
        try {
            let restApi = await this.makeConnection(headerConfigProps);
            let uniqueTCsdetails = await this.getAllTCByProject(restApi, projectName);
            if (uniqueTCsdetails) {
                await this.postTCResultCollectionForRepeatedTC(restApi, uniqueTCsdetails, headerConfigProps["apikey"], proxy);
            }
        } catch (error) {
            console.debug("Error while cleaning up using project Name: " + error.message)
        }
    }

    async getAllTCByProject(restApi, projectName) {
        let qr = queryUtils.where("Project.Name", "=", projectName);
        let allUniqueGroupedTCs = {};

        //let qr = queryUtils.where("Name", "in", ["KraftRallyTestAssertionNewFolder"]);
        let reqedFoldersInProj = await restApi.query({
            type: 'TestFolder', //the type to query
            start: 1, //the 1-based start index, defaults to 1
            pageSize: 200, //the page size (1-200, defaults to 200)
            limit: Infinity, //the maximum number of results to return- enables auto paging
            fetch: ['Name', 'TestCases', 'Children'], //the fields to retrieve
            scope: {},
            query: qr
        });
        if (reqedFoldersInProj.Errors.length <= 0) {
            let results = reqedFoldersInProj.Results;
            if (results.length > 0) {
                for (let testfolder of results) {
                    let qr = queryUtils.where("TestFolder", "=", testfolder._ref);
                    let requestedTestCases = await restApi.query({
                        type: 'TestCase', //the type to query
                        start: 1, //the 1-based start index, defaults to 1
                        pageSize: 200, //the page size (1-200, defaults to 200)
                        limit: Infinity, //the maximum number of results to return- enables auto paging
                        fetch: ["Method", "Package", "Priority", "Risk", "Type", "WorkProduct", "FormattedId", "Name"], //the fields to retrieve
                        scope: {},
                        query: qr
                    });
                    if (requestedTestCases.Errors.length > 0) {
                        throw new Error("Error Finding testcase details : " + requestedTestCases.Errors.join("\n"));
                    }
                    if (requestedTestCases.Results.length > 0) {
                        let uniqueGroups = await this.findRepeatedTestCasesNames(requestedTestCases.Results);
                        allUniqueGroupedTCs = { ...allUniqueGroupedTCs, ...uniqueGroups };
                    }
                }
            } else {
                console.debug("No Foldersin given project: " + projectName);
            }
        } else {
            throw new Error("Error while fetching folders: " + reqedFoldersInProj.Errors.join('\n'));
        }
        return allUniqueGroupedTCs;
    }

    async findRepeatedTestCasesNames(allTestCases) {
        const groups = allTestCases.reduce((groups, item) => {
            let group = (groups[item.Name] || []);
            group.push(item);
            groups[item.Name] = group;
            return groups;
        }, {});
        let uniqueGroups = {};
        let uniqueTCNames = Object.keys(groups);

        for (let tcName of uniqueTCNames) {
            if (groups[tcName].length > 1)
                uniqueGroups[tcName] = groups[tcName]
        }
        return uniqueGroups;
    }

    async postTCResultCollectionForRepeatedTC(restApi, uniqueGroups, apikey, proxy) {
        let repeatedTCNames = Object.keys(uniqueGroups);
        let tcResultCollectRefactor = [];
        let toDelTCCollection = [];
        for (let tcName of repeatedTCNames) {
            let tcResults = await this.getTCResultDet(restApi, tcName);

            if (tcResults.Errors.length > 0)
                throw new Error("Error fetching testcaseResults:" + tcResults.Errors.join("\n"));
            else if (tcResults.Results.length > 0) {
                let onlyNum = uniqueGroups[tcName].map(value => parseInt(value.FormattedID.substring(value.FormattedID.indexOf("TC") + 2)));
                let latestTCDet = uniqueGroups[tcName].find((tc) => { return tc.FormattedID === "TC" + Math.max(...onlyNum).toString() });
                uniqueGroups[tcName].splice(uniqueGroups[tcName].findIndex(a => a.FormattedID === latestTCDet.FormattedID), 1)
                for (let tcRept of uniqueGroups[tcName]) {
                    let tcDelEntry = {
                        "Entry": {
                            "Path": "",
                            "Method": "DELETE",
                        }
                    }
                    tcDelEntry.Entry.Path = tcRept._ref.slice(tcRept._ref.indexOf('/testcase'))
                    toDelTCCollection.push(tcDelEntry)
                }
                //   let firstTCDet = groups[tcName].find((tc) => { return tc.FormattedID === "TC" + Math.min(...onlyNum).toString() });
                let filterTestSteps = tcResults.Results.filter((result) => { return result.TestCase._ref != latestTCDet._ref });
                for (let teststep of tcResults.Results) {
                    let tcResultDet = {};
                    tcResultDet.TestCase = latestTCDet._ref;
                    tcResultDet.Project = teststep.Project._ref;
                    tcResultDet.Tester = teststep.Tester._ref;
                    tcResultDet.Build = teststep.Build;
                    tcResultDet.Verdict = teststep.Verdict;
                    tcResultDet.Duration = teststep.Duration;
                    tcResultDet.Date = teststep.Date;
                    tcResultDet.Method = teststep.Method;
                    if (teststep.TestSet) {
                        tcResultDet.TestSet = teststep.TestSet._ref;
                    }
                    tcResultDet.Notes = teststep.Notes;
                    let tcResultEntry = await this.createTcResultEntry(tcResultDet);
                    tcResultCollectRefactor.push(tcResultEntry);

                }
            }
        }
        await RallyBatchAPIController.performBatchOperation(tcResultCollectRefactor, apikey, proxy);
        await RallyBatchAPIController.performBatchOperation(toDelTCCollection, apikey, proxy,"SuccessFully deleted Tcs", "Error while deleting TC")
    }

    async getTCResultDet(restApi, arrOfTCs) {
        let qr
        qr = queryUtils.where("Testcase.Name", "=", arrOfTCs);
        // for(let index = 1 ; index < arrOfTCs.length ; index++) 
        //     qr.or("Testcase", "=", arrOfTCs[index]["_ref"])
        return restApi.query({
            type: 'Testcaseresult', //the type to query
            start: 1, //the 1-based start index, defaults to 1
            pageSize: 200, //the page size (1-200, defaults to 200)
            limit: Infinity, //the maximum number of results to return- enables auto paging
            fetch: ["Build", "Date", "Duration", "Notes", "Project", "Tester", "testset", "verdict", "TestCase", "Name"], //the fields to retrieve
            scope: {},
            query: qr
        });
    }

    async createDummyBulkTestacases(restApi, apikey, proxy) {
        let testCaseEntries = [];
        let qr = queryUtils.where("Name", "=", 'Tools and Performance Engineering');
        let projectRef, folderRef;
        let reqedProjects = await restApi.query({
            type: 'Project', //the type to query
            start: 1, //the 1-based start index, defaults to 1
            pageSize: 200, //the page size (1-200, defaults to 200)
            limit: Infinity, //the maximum number of results to return- enables auto paging
            fetch: ['Project'], //the fields to retrieve
            scope: {},
            query: qr
        });
        if ( reqedProjects.Results.length > 0 ) {
            projectRef = reqedProjects.Results[0]._ref;
        }
        let qr1 = queryUtils.where("Name", "=", 'DummyCleanupFolder2');
        qr1 = qr1.and("Project", "=", projectRef)
        let reqedFolder = await restApi.query({
            type: 'TestFolder', //the type to query
            start: 1, //the 1-based start index, defaults to 1
            pageSize: 200, //the page size (1-200, defaults to 200)
            limit: Infinity, //the maximum number of results to return- enables auto paging
            fetch: ['TestFolder'], //the fields to retrieve
            scope: {},
            query: qr1
        });
        if ( reqedFolder.Results.length > 0 ) {
            folderRef = reqedFolder.Results[0]._ref;
        }
        for(let i=0; i<=10010; i++) {
            let testcaseName = "Sample Testcase for testing cleanup " + i;
            let testcaseEntry = {
                "Entry" : {
                    "Path": "/testcase/create",
                    "Method": "POST",
                    "Body": {
                        "testcase" : {
                            "Name": testcaseName,
                            "TestFolder": folderRef   
                        }                 
                    }
                }
            }
            testcaseEntry.Entry.Body.Name = testcaseName;
            testCaseEntries.push(testcaseEntry);
        }
        let allTCRefs = await RallyBatchAPIController.postTestCases(testCaseEntries, apikey, proxy, true);
        return allTCRefs;
    } 

    async createTestCaseUpdateEntries(tcCollection) {
        for ( let tc of tcCollection ) {
            let entry = {
                "Entry": {
                    "Path" : tc._ref,
                    "Method" : "PUT",
                    "Body" : {
                        "testcase" : {
                            "TestFolder": "https://rally1.rallydev.com/slm/webservice/v2.0/testfolder/635042010939"
                        }
                    }
                }
            }

        }
    }

    /* async createDeleteEntriesByIds(tcIds, createDelEntriesAPI, queryUtils) {
        try {
            let tcIdsCollection = tcIds.split(",");
            let tcRefEntries = []
            const noOfTcToGetPerBatch = 500;
            let noOfIterToGetAllTc;
            let notFetchTcIds = [];
            if (tcIdsCollection.length > noOfTcToGetPerBatch) { noOfIterToGetAllTc = (tcIdsCollection.length % noOfTcToGetPerBatch) === 0 ? tcIdsCollection.length / noOfTcToGetPerBatch : Math.floor(tcIdsCollection.length / noOfTcToGetPerBatch) + 1 }
            else noOfIterToGetAllTc = 1;
            for (let iterationCount = 0; iterationCount < noOfIterToGetAllTc; iterationCount++) {
                let tcIdsToFetchPerItr = tcIdsCollection.splice(0, noOfTcToGetPerBatch) ;

                let qr = queryUtils.where("FormattedID", "in", tcIdsToFetchPerItr)
                let reqedTestcase = await createDelEntriesAPI.query({
                    type: 'Testcase', //the type to query
                    start: 1, //the 1-based start index, defaults to 1
                    pageSize: 200, //the page size (1-200, defaults to 200)
                    limit: Infinity, //the maximum number of results to return- enables auto paging
                    fetch: ['FormattedId'], //the fields to retrieve
                    scope: {},
                    query: qr
                });
                if (reqedTestcase.Errors.length <= 0) {
                    let fetchedIds = reqedTestcase.Results.map(res => res.FormattedID);
                    notFetchTcIds.push(...(tcIdsToFetchPerItr.filter(value => !(fetchedIds.includes(value)))));
                    tcRefEntries.push(...(reqedTestcase.Results.map((tcDet) => {
                        return {
                            "Entry": {
                                "Path": tcDet._ref.slice(tcDet._ref.indexOf('/testcase')),
                                "Method": "DELETE"
                            }
                        }
                    })))
                }
            }
            if (notFetchTcIds.length > 0)
                console.debug("Not fetched Tc details of Ids: " + JSON.stringify(notFetchTcIds));
            return tcRefEntries;
        } catch (ex) {
            console.error("Error Occured while fetching tc details : " + ex.message)
        }
    } */
}
module.exports = new RallyCleanup();