/**
 * Created by 212556710 on 4/28/16.
 */
 module.exports = function JsonOutputHook() {
    var Cucumber = require('cucumber');
    require("dotenv").config();

    var Logger = require('./Logger.js');
    var JsonFormatter = Cucumber.Listener.JsonFormatter();
    var json_rally = '';
    var fs = require('fs');
    var dir = './Reports/json_report';
    if (!fs.existsSync(dir)) {
        fs.mkdirSync(dir);
    }

    var imageCount = 0;
    this.registerListener(JsonFormatter);
    JsonFormatter.log = function (jsonold) {
        json_rally = jsonold;
    };

    this.After(function (scenario, callback) {
        console.log('after scenario');
        if (scenario.isFailed()) {
            Logger.info("Results: Failed");
            browser.takeScreenshot().then(function (png) {
                imageCount += 1;
                var decodedImage = new Buffer(png, "base64").toString("binary");
                fs.writeFile("Reports/image" + imageCount + ".png", decodedImage, 'binary', function (err) {
                    scenario.attach(decodedImage, "image/png");
                    console.log("Attached");
                    callback()
                })
            });
        } else {
            Logger.info("Results: Passed");
            callback();
        }
    });

    this.registerHandler('BeforeScenario', function (event, callback) {
        var scenario = event.getPayloadItem('scenario');
        Logger.info("|--------------------------------------------------------|");
        Logger.info("Test " + scenario.getName() + " has started");
        Logger.info("|--------------------------------------------------------|\r\n");
        callback();
    });

    this.registerHandler('AfterScenario', function (event, callback) {
        var scenario = event.getPayloadItem('scenario');
        Logger.info("|--------------------------------------------------------|");
        Logger.info("Test " + scenario.getName() + " has ended");
        Logger.info("|--------------------------------------------------------|\r\n\r\n");
        callback();
    });

    this.registerHandler('AfterFeatures', function (features, callback) {
        if (json_rally == '') {
            //spin for loop until json rally finished
            var date = new Date();
            while (new Date() - date < 1500) {
                //spin for 1.5 sec
            }
        }
        
        if (process.env.postToRally == "true" || process.env.posttorally == true) {
            var adaptors = require('./adaptors.js');
            const ParseAndPostTestData = require('./ParseTestData');

            Logger.debug("Started Rally Posting")
            let headers =  {
                "framework": "Proui",
                "content-type": "application/json",
                "updatedtestcases" : process.env.updatedTestcases,
                "userstory": process.env.user_story ? process.env.user_story : process.env.userstory,
                "testsetid": process.env.testset ? process.env.testset : process.env.testsetId,
                "build": process.env.build_no? process.env.build_no : process.env.build,
                "method": process.env.method,
                "testtype": process.env.type? process.env.type : process.env.testtype,
                "tags": process.env.tag ? process.env.tag : process.env.tags,
                "rallyworkspace": process.env.rallyworkspace,
                "folderpath": process.env.rallyrootfolder ? process.env.rallyrootfolder : process.env.folderpath,
                "projectname": process.env.project? process.env.project : process.env.projectname,
                "apikey": process.env.apiKey,
                "rallyproxy": process.env.rallyProxy,
                "overwrite_fp": process.env.overwrite_RF? process.env.overwrite_RF : process.env.overwrite_FP,
                "overwrite_pn": process.env.overwrite_PN,
                "overwrite_ts": process.env.overwrite_TS,
                "overwrite_us": process.env.overwrite_US,
                "overwrite_build": process.env.overwrite_build,
                "overwrite_tags": process.env.overwrite_tags,
                "overwrite_method": process.env.overwrite_method,
                "overwrite_testtype": process.env.overwrite_testtype,
                "updatetcr": process.env.overwrite_result ? process.env.overwrite_result : process.env.updateTCR,
                "createnewtestcases": process.env.createNewTestCases,
                "generatetrends": process.env.generateTrendsJson ? process.env.generateTrendsJson : process.env.generateTrends,
                "trendsmodule": process.env.Trends_Module? process.env.Trends_Module : process.env.TrendsModule,
                "trendsenv": process.env.Trends_Env? process.env.Trends_Env : process.env.TrendsEnv,
                "trendsjobname": process.env.Trends_JobName ? process.env.Trends_JobName : process.env.TrendsJobName,
                "awss3proxy": process.env.AWSS3Proxy,
                "rallyusername": process.env.RallyUserName
    

            };
            headers = JSON.parse(JSON.stringify(headers).replace(/"\s+|\s+"/g, '"'))
            let testResults = adaptors.convertOldUIResults(JSON.parse(json_rally), headers.overwrite_us, headers.userstory)

            let parseAndPostTestDataObj = new ParseAndPostTestData();
            // calling the main function of rally posting 
            parseAndPostTestDataObj.postResults(headers, testResults).then(function (result) {
                let respMsg = '';
                if (result.errors) {
                    if (result.postingError)
                      respMsg = "Failed to post Test Results to rally\nLogs:\n=====\n";
                    else if (result.trendsPostingError) {
                      respMsg = "Successfully posted Test Results to rally.\n"
                      respMsg += "Failed to generate trends\nLogs:\n=====\n";
                    }
                    console.error(respMsg + result.message + "\nPosting Processing done. " + new Date());
                    callback();
                } else {
                    respMsg += "Successfully posted Test Results to Rally\nLogs:\n====\n"
                    console.log(respMsg + result.message + "\nPosting Processing done. " + new Date());
                    callback();     
                }

            }).catch(function (error) {
                console.error("Errors in Posting \n----------------------\nResponse Logs: \n" + error.message + "\nstack : " + error.stack);
                callback();
            });

        } else {
            Logger.info("Not posting results to rally as postToRally is not set to true")
            console.log("Not posting results to rally as postToRally is not set to true");
            callback();
        }
    });
};
