let RallyTestResult = class  {

	#mWorkspaceName = null;
	#mProjectName = null;
	#mTestFolderPath = null;

	#mFeatureName = null;
	#mFeatureNamePath = null;
	#mTestName = null;
	#mTestCaseID = null;
	#mDescription = null;
	#mBuild = null;
	#mTestStatus = null;
	#mTestType = null;
	#mNotes = null;
	#mDurationMS = null;
	#mStartedAt = null;
	#mMessage = null;
	#mStackTrace = null;
	#mObjective = null;
	#mPreConditions = null;
	#mPostConditions = null;
	#mPriority = null;
	#mRisk = null;
	#mUserStory = null;

	// #mTCStepInput = null;
	#mTCStepExpResult = null;
	#mDefectId = null;
	#mMethod = null;
	#mTags = null;
	#mTestSetId = null;

	#mSpecialChars = "~`!@#$%^&*()-_=+\\|[{]};:'\",<.>/?";
	#attributeList = null;
	invalidDataCollection = null;

	constructor() {
		this.#attributeList = new Array();
		this.invalidDataCollection = new Array();
		this.#attributeList.push("workspacename");
		this.#attributeList.push("folderpath");
		this.#attributeList.push("projectname");
		this.#attributeList.push("testcaseid");
		this.#attributeList.push("testsetid");
		this.#attributeList.push("userstory");
		this.#attributeList.push("description");
		this.#attributeList.push("preconditions");
		this.#attributeList.push("postconditions");
		this.#attributeList.push("objective");
		// this.#attributeList.push("tcstepinput");
		// this.#attributeList.push("tcstepexpresult");
		this.#attributeList.push("method");
		this.#attributeList.push("build");
		this.#attributeList.push("defectid");
		this.#attributeList.push("risk");
		this.#attributeList.push("priority");
		this.#attributeList.push("tags");
		this.#attributeList.push("testtype");
	}

	async verifySpecialCharacter(input, message) {
		let pass = true;
		for (let i = 0; i < this.#mSpecialChars.length; i++) {
			if (input.indexOf(this.#mSpecialChars.charAt(i)) > -1) {
				pass = false;
				break;
			}
		}
		if (!pass)
			this.invalidDataCollection.push(message);
	}

	async verifyInvalidAttributes(attributes) {
		for (let attributeName of attributes) {
			if (!this.#attributeList.includes(attributeName))
				this.invalidDataCollection
					.push("Invalid '" + attributeName + "' attribute in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath() + "'. ");
		}
	}

	async verifyInvalidAttributeValues(attributeName, attributeValue) {
		if (attributeValue == null ||((typeof attributeValue === 'string' || attributeValue instanceof String) && ( attributeValue.length <= 0 || attributeValue.trim().length <= 0
			|| attributeValue.length > 32768))) {
			this.invalidDataCollection.push("Invalid '" + attributeName + "' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath()
				+ "'. '" + attributeName + "' attribute value should not be empty or greater than 32768 character.");
		}
	}

	async getSpecialChars() {
		return this.#mSpecialChars;
	}

	async getWorkspaceName() {
		return this.#mWorkspaceName;
	}

	async setWorkspaceName(mWorkspaceName) {
		this.#mWorkspaceName = mWorkspaceName;
		this.verifyInvalidAttributeValues("WorkspaceName", this.#mWorkspaceName);
	}

	async getProjectName() {
		return this.#mProjectName;
	}

	async setProjectName(mProjectName) {
		this.#mProjectName = mProjectName;
		this.verifyInvalidAttributeValues("ProjectName", this.#mProjectName);
	}

	async getTestFolderPath() {
		return this.#mTestFolderPath;
	}

	async setTestFolderPath(mTestFolderPath) {
		let folderNames = mTestFolderPath.split("/");
		let index = 0;
		for (let folderName of folderNames)
			folderNames[index++] = folderName.trim();
		let trimmedTestFolderPath = folderNames.join("/");
		this.#mTestFolderPath = trimmedTestFolderPath;
		this.verifyInvalidAttributeValues("FolderPath", this.#mTestFolderPath);
	}

	/**
	 *  
	 * 
	 * @return
	 */
	async getTestCaseID() {
		return this.#mTestCaseID;
	}

	/**
	 *  
	 * 
	 * @param mTestCaseID
	 */
	async setTestCaseId(mTestCaseID) {
		this.#mTestCaseID = mTestCaseID;
		this.verifyInvalidAttributeValues("TestCaseID", this.#mTestCaseID);
	}

	async getFeatureName() {
		return this.#mFeatureName;
	}

	async setFeatureName(mFeatureName) {
		this.#mFeatureName = mFeatureName;
	}

	async getFeatureNamePath() {
		return this.#mFeatureNamePath;
	}

	async setFeatureNamePath(mFeatureNamePath) {
		this.#mFeatureNamePath = mFeatureNamePath;
	}

	async getTestSetId() {
		return this.#mTestSetId;
	}

	async setTestSetId(mTestSetId) {
		this.#mTestSetId = mTestSetId;
		this.verifyInvalidAttributeValues("TestSetId", this.#mTestSetId);
	}

	/**
	 *  
	 * 
	 * @return
	 */
	async getUserStory() {
		return this.#mUserStory;
	}

	/**
	 *  
	 * 
	 * @param mUserStory
	 */
	async setUserStory(mUserStory) {
		this.#mUserStory = mUserStory;
		this.verifyInvalidAttributeValues("UserStory", this.#mUserStory);
	}

	/**
	 * The Test Case Name to which this result belongs
	 */
	async getTestName() {
		return this.#mTestName;
	}

	/**
	 * The Test Case Name to which this result belongs
	 */
	async setTestName(mTestName) {
		this.#mTestName = mTestName;
	}

	/**
	 * The description for the test case
	 * 
	 * @return
	 */
	async getDescription() {
		return this.#mDescription;
	}

	/**
	 * The description for the test case  
	 */
	async setDescription(mDescription) {
		this.#mDescription = mDescription;
		this.verifyInvalidAttributeValues("Description", this.#mDescription);
	}

	/**
	 *  
	 * 
	 * @return
	 */
	async getBuild() {

		//		if(#mBuild == null || #mBuild.isEmpty()){
		//			let rnd = (let)(Math.random()*100);
		//			#mBuild = .valueOf(rnd);
		//		}
		//		
		//		
		return this.#mBuild;
	}

	/**
	 * The defects of the test. 
	 */
	async getDefectId() {
		return this.#mDefectId;
	}

	/**
	 * The defect(s) for the test case  
	 */
	async setDefectId(mDefectId) {
		this.#mDefectId = mDefectId;
		this.verifyInvalidAttributeValues("DefectId", this.#mDefectId);
	}

	async setBuild(mBuild) {
		this.#mBuild = mBuild;
		this.verifyInvalidAttributeValues("Build", this.#mBuild);
	}

	/**
	 * The status of the test. 
	 * 
	 * @return One of "Blocked", "Error", "Fail", "Inconclusive", "Pass"
	 */
	async getTestStatus() {
		return this.#mTestStatus;
	}

	/**
	 * The status of the test. 
	 * 
	 * @return One of "Blocked", "Error", "Fail", "Inconclusive", "Pass"
	 * @throws IllegalArgumentException
	 */
	async setTestStatus(mTestStatus) {

		// switch (mTestStatus.toUpperCase()) {
		// 	case "FAILED":
		// 		rResult.setTestStatus("Fail");
		// 		break;
		// 	case "PASSED":
		// 		rResult.setTestStatus("Pass");
		// 		break;
		// 	case "SKIP":
		// 		rResult.setTestStatus("Blocked");
		// 		break;
		// 	case "PENDING":
		// 		rResult.setTestStatus("Inconclusive");
		// 		break;
		// 	case "UNDEFINED":
		// 		rResult.setTestStatus("Error");
		// 		break;
		// 	default:
		// 		this.invalidDataCollection.push("Invalid 'teststatus' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath()
		// 			+ "'. Allowed 'teststatus' attribute values are 'blocked', 'error', 'fail','inconclusive' and 'pass'.");
		// }

		this.#mTestStatus = mTestStatus;
	}
	

	async getTestType() {
		return this.#mTestType;
	}

	/**
	 *  
	 * 
	 * @param mTestType Allowed values are: "Acceptance", "Functional",
	 *                  "Performance", "Regression", "Usability", "User leterface","Obsolete"
	 */
	async setTestType(mTestType) {
		switch (mTestType.trim().toLowerCase()) {
			case "acceptance":
				this.#mTestType = "Acceptance";
				break;
			case "functional":
				this.#mTestType = "Functional";
				break;
			case "performance":
				this.#mTestType = "Performance";
				break;
			case "regression":
				this.#mTestType = "Regression";
				break;
			case "usability":
				this.#mTestType = "Usability";
				break;
			case "user interface":
				this.#mTestType = "User Interface";
				break;
			case "obsolete":
				this.#mTestType = "Obsolete";
				break;
			default:
				this.invalidDataCollection.push("Invalid 'testtype' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath() + "'. ");
		}
	}

	async getNotes() {
		return this.#mNotes;
	}

	async setNotes(mNotes) {
		this.#mNotes = mNotes;
	}

	async getDurationMS() {
		return this.#mDurationMS;
	}

	async setDurationMS(mDurationMS) {
		this.#mDurationMS = mDurationMS;
	}

	async getObjective() {
		return this.#mObjective;
	}

	async setObjective(mObjective) {
		this.#mObjective = mObjective;
		this.verifyInvalidAttributeValues("Objective", this.#mObjective);
	}

	async getPriority() {
		return this.#mPriority;
	}

	async setPriority(mPriority) {
		switch (mPriority) {
			case "Useful":
			case "Important":
			case "Critical":
				break;
			default:
				this.invalidDataCollection.push("Invalid 'priority' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath()
					+ "'. Allowed 'priority' attribute values are 'Useful', 'Important' and 'Critical' and is case-sensitive");
		}
		this.#mPriority = mPriority;
	}

	async getRisk() {
		return this.#mRisk;
	}

	async setRisk(mRisk) {
		switch (mRisk) {
			case "Low":
			case "Medium":
			case "High":
				break;
			default:
				this.invalidDataCollection.push("Invalid 'risk' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath()
					+ "'. Allowed 'risk' attribute values are 'Low', 'Medium' and 'High' and is case-sensitve");
		}
		this.#mRisk = mRisk;
	}

	async getPreConditions() {
		return this.#mPreConditions;
	}

	async setPreConditions(mPreConditions) {
		this.#mPreConditions = mPreConditions;
		this.verifyInvalidAttributeValues("PreConditions", this.#mPreConditions);
	}

	async getPostConditions() {
		return this.#mPostConditions;
	}

	async setPostConditions(mPostConditions) {
		this.#mPostConditions = mPostConditions;
		this.verifyInvalidAttributeValues("PostConditions", this.#mPostConditions);
	}

	async getStackTrace() {
		return this.#mStackTrace;
	}

	async setStackTrace(mStackTrace) {
		this.#mStackTrace = mStackTrace;
	}

	// async getTCStepInput() {
	// 	return this.#mTCStepInput;
	// }

	// async setTCStepInput(mTCStepInput) {
	// 	this.#mTCStepInput = mTCStepInput;
	// 	this.verifyInvalidAttributeValues("TCStepInput", this.#mTCStepInput);
	// }

	async getTCStepExpResult() {
		return this.#mTCStepExpResult;
	}

	async setTCStepExpResult(mTCStepExpResult) {
		this.#mTCStepExpResult = mTCStepExpResult;
		this.verifyInvalidAttributeValues("TCStepExpResult", this.#mTCStepExpResult);
	}

	async setMethod(mMethod) {
		switch (mMethod.trim().toLowerCase()) {
			case "manual":
				this.#mMethod = "Manual";
				break;
			case "automated":
				this.#mMethod = "Automated";
				break;
			default:
				this.invalidDataCollection.push("Invalid 'Method' attribute value in scenario '" + await this.getTestName() + "' of '" + await this.getFeatureNamePath()+ "'.");
		}
		this.verifyInvalidAttributeValues("Method", this.#mMethod);
	}

	async getExecutionType() {
		return this.#mMethod;
	}

	async getTags() {
		return this.#mTags;
	}

	async setTags(mTags) {
		this.#mTags = mTags;
		this.verifyInvalidAttributeValues("Tags", this.#mTags);
	}

}
module.exports = RallyTestResult;